/*
 *  This file is part of x48, an emulator of the HP-48sx Calculator.
 *  Copyright (C) 1994  Eddie C. Dost  (ecd@dressler.de)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* $Log: debugger.c,v $
 * Revision 1.4  1994/10/09  20:29:47  ecd
 * start of disassembler implementation.
 *
 * Revision 1.3  1994/10/06  16:30:05  ecd
 * added refresh_display()
 *
 * Revision 1.2  1994/10/05  08:36:44  ecd
 * more functions
 *
 * Revision 1.1  1994/10/04  15:12:21  ecd
 * Initial revision
 *
 *
 * $Id: debugger.c,v 1.4 1994/10/09 20:29:47 ecd Exp ecd $
 */

#include "global.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#ifdef SUNOS
#include <memory.h>
#endif
#include <X11/Xlib.h>

#include "hp48.h"
#include "device.h"
#include "timer.h"
#include "x48_x11.h"
#include "debugger.h"
#include "disasm.h"

extern char *readline    __ProtoType__((const char *));
extern void  add_history __ProtoType__((char *));

#define MAX_ARGS 16

int enter_debugger = 0;

static int continue_flag;
static char instr[100];

static void do_continue	__ProtoType__((int, char **));
static void do_exit	__ProtoType__((int, char **));
static void do_go	__ProtoType__((int, char **));
static void do_help	__ProtoType__((int, char **));
static void do_load	__ProtoType__((int, char **));
static void do_mode	__ProtoType__((int, char **));
static void do_quit	__ProtoType__((int, char **));
static void do_regs	__ProtoType__((int, char **));
static void do_save	__ProtoType__((int, char **));
static void do_stat	__ProtoType__((int, char **));
static void do_step	__ProtoType__((int, char **));
static void do_reset	__ProtoType__((int, char **));
static void do_rstk	__ProtoType__((int, char **));

struct cmd {
  char *name;
  void (*func) __ProtoType__((int, char**));
  char *help;
} cmd_tbl[] = {
  { "cont",	do_continue,
    "cont                       Continue execution" },
  { "c",	do_continue,	0 },

  { "exit",	do_exit,
    "exit                       Exit the emulator without saving" },

  { "go",	do_go,
    "go hexvalue                Set PC to address `hexvalue\'" },

  { "help",	do_help,
    "help                       Display this information" },
  { "h",	do_help,	0 },
  { "?",	do_help,	0 },

  { "load",	do_load,
    "load                       Load emulator-state from files" },

  { "mode",	do_mode,
    "mode [hp | alonzo]         Show or set disassembler mode" },

  { "quit",	do_quit,
    "quit                       Exit the emulator with saving it\'s state" },
  { "q",	do_quit,	0 },

  { "reg",	do_regs,
    "reg [register [hexvalue]]  Display or set register value" },
  { "r",	do_regs,	0 },

  { "reset",	do_reset,
    "reset                      Set the HP48\'s PC to ZERO" },

  { "save",	do_save,
    "save                       Save emulator-state to files" },

  { "stat",	do_stat,
    "stat                       Display statistics for the emulator" },

  { "step",	do_step,
    "step                       Step one Instruction" },
  { "s",	do_step,	0 },

  { "where",	do_rstk,
    "where                      Show ML return stack" },

  { 0,		0,		0 }
};

static inline void
#ifdef __FunctionProto__
str_to_upper(char *arg)
#else
str_to_upper(arg)
char *arg;
#endif
{
  int i;

  for (i = 0; i < strlen(arg); i++) {
    if ('a' <= arg[i] && arg[i] <= 'z') {
      arg[i] = (char)((int)arg[i] - (int)'a' + (int)'A');
    }
  }
}

static int
#ifdef __FunctionProto__
decode_20(word_20 *addr, char *arg)
#else
decode_20(addr, arg)
word_20 *addr;
char *arg;
#endif
{
  int i;

  if (arg == (char *)0) {
    printf("Command requires an argument.\n");
    return 0;
  }

  *addr = 0;
  for (i = 0; i < strlen(arg); i++) {
    *addr <<= 4;
    if ('0' <= arg[i] && arg[i] <= '9') {
      *addr |= ((int)arg[i] - (int)'0');
    } else if ('A' <= arg[i] && arg[i] <= 'F') {
      *addr |= ((int)arg[i] - (int)'A' + 10);
    } else {
      *addr = 0;
      printf("Not a number: %s.\n", arg);
      return 0;
    }
    *addr &= 0xfffff;
  }
  return 1;
}

static int
#ifdef __FunctionProto__
decode_64(word_64 *addr, char *arg)
#else
decode_64(addr, arg)
word_64 *addr;
char *arg;
#endif
{
  int i;

  if (arg == (char *)0) {
    printf("Command requires an argument.\n");
    return 0;
  }

  addr->lo = addr->hi = 0;
  for (i = 0; i < strlen(arg); i++) {
    addr->hi <<= 4;
    addr->hi |= ((addr->lo >> 28) & 0x0f);
    addr->lo <<= 4;
    if ('0' <= arg[i] && arg[i] <= '9') {
      addr->lo |= ((int)arg[i] - (int)'0');
    } else if ('A' <= arg[i] && arg[i] <= 'F') {
      addr->lo |= ((int)arg[i] - (int)'A' + 10);
    } else {
      addr->hi = addr->lo = 0;
      printf("Not a number: %s.\n", arg);
      return 0;
    }
  }
  return 1;
}

static char *
#ifdef __FunctionProto__
str_nibbles(word_20 addr, int n)
#else
str_nibbles(addr, n)
word_20 addr;
int n;
#endif
{
  static char str[1025];
  char *cp;
  int i;

  if (n > 1024) {
    str[0] = '\0';
    return str;
  }

  for (cp = str, i = 0; i < n; i++) {
    sprintf(cp, "%.1X", read_nibble(addr + i));
    cp++;
  }
  *cp = '\0';

  return str;
}

static int
#ifdef __FunctionProto__
confirm(const char *prompt)
#else
confirm(prompt)
const char *prompt;
#endif
{
  char ans[80];

  printf("%s (y or n) ", prompt);
  fflush(stdout);
  fgets(ans, sizeof(ans), stdin);
  while (ans[0] != 'y' && ans[0] != 'Y' && ans[0] != 'n' && ans[0] != 'N') {
    printf("Please answer y or n.\n");
    printf("%s (y or n) ", prompt);
    fflush(stdout);
    fgets(ans, sizeof(ans), stdin);
  }
  if (ans[0] == 'y' || ans[0] == 'Y') {
    return 1;
  } else {
    printf("Not confirmed.\n");
    return 0;
  }
}

static void
#ifdef __FunctionProto__
do_continue(int argc, char **argv)
#else
do_continue(argc, argv)
int argc;
char *argv;
#endif
{
  continue_flag = 1;
}

static void
#ifdef __FunctionProto__
do_exit(int argc, char **argv)
#else
do_exit(argc, argv)
int argc;
char *argv;
#endif
{
  if (confirm("Exit the emulator WITHOUT saving it\'s state?")) {
    printf("Exit.\n");
    exit(0);
  }
}

static void
#ifdef __FunctionProto__
do_go(int argc, char **argv)
#else
do_go(argc, argv)
int argc;
char *argv;
#endif
{
  word_20 addr;

  str_to_upper(argv[1]);
  if (decode_20(&addr, argv[1])) {
    saturn.PC = addr;
    enter_debugger &= ~ILLEGAL_INSTRUCTION;
  }
}

static void
#ifdef __FunctionProto__
do_help(int argc, char **argv)
#else
do_help(argc, argv)
int argc;
char *argv;
#endif
{
  int i;

  for (i = 0; cmd_tbl[i].name; i++) {
    if (cmd_tbl[i].help) {
      printf("%s.\n", cmd_tbl[i].help);
    }
  }
}

static void
#ifdef __FunctionProto__
do_load(int argc, char **argv)
#else
do_load(argc, argv)
int argc;
char *argv;
#endif
{
  saturn_t tmp_saturn;
  device_t tmp_device;

  if (confirm("Load emulator-state from files?")) {
    memcpy(&tmp_saturn, &saturn, sizeof(saturn));
    memcpy(&tmp_device, &device, sizeof(device));
    memset(&saturn, 0, sizeof(saturn));
    if (read_files()) {
      printf("Loading done.\n");
      enter_debugger &= ~ILLEGAL_INSTRUCTION;
      if (tmp_saturn.rom) {
        free(tmp_saturn.rom);
      }
      if (tmp_saturn.ram) {
        free(tmp_saturn.ram);
      }
      if (tmp_saturn.port1) {
        free(tmp_saturn.port1);
      }
      if (tmp_saturn.port2) {
        free(tmp_saturn.port2);
      }
      init_display();
      update_display();
#ifdef USE_SHM
      if (disp.display_update) refresh_display();
#endif
    } else {
      printf("Loading emulator-state from files failed.\n");
      if (saturn.rom) {
        free(saturn.rom);
      }
      if (saturn.ram) {
        free(saturn.ram);
      }
      if (saturn.port1) {
        free(saturn.port1);
      }
      if (saturn.port2) {
        free(saturn.port2);
      }
      memcpy(&saturn, &tmp_saturn, sizeof(saturn));
      memcpy(&device, &tmp_device, sizeof(device));
    }
  }
}

static void
#ifdef __FunctionProto__
do_mode(int argc, char **argv)
#else
do_mode(argc, argv)
int argc;
char *argv;
#endif
{
  if (argc < 2) {
    printf("Disassembler uses %s mnemonics.\n",
           disassembler_mode == HP_MNEMONICS ? "HP" : "Alonzo");
  } else {
    str_to_upper(argv[1]);
    if (!strcmp("HP", argv[1])) {
      disassembler_mode = HP_MNEMONICS;
    } else if (!strcmp("ALONZO", argv[1])) {
      disassembler_mode = ALONZO_MNEMONICS;
    } else {
      printf("Unknown disassembler mode %s. Try \"help\".\n", argv[1]);
    }
  }
}

static void
#ifdef __FunctionProto__
do_quit(int argc, char **argv)
#else
do_quit(argc, argv)
int argc;
char *argv;
#endif
{
  if (confirm("Quit the emulator and save it\'s state?")) {
    printf("Exit.\n");
    exit_emulator();
    exit(0);
  }
}

static void
#ifdef __FunctionProto__
set_reg(word_64 val, int n, unsigned char *r)
#else
set_reg(val, n, r)
word_64 val;
int n;
unsigned char *r;
#endif
{
  int i;

  for (i = 0; i < n; i++) {
    if (i < 8)
      r[i] = (unsigned char)((val.lo & (0xf << (4*i))) >> (4*i));
    else
      r[i] = (unsigned char)((val.hi & (0xf << (4*(i-8)))) >> (4*(i-8)));
  }
}

static void
#ifdef __FunctionProto__
dump_reg(const char *reg, int n, unsigned char *r)
#else
dump_reg(reg, n, r)
const char *reg;
int n;
unsigned char *r;
#endif
{
  int i;

  printf("%s:\t", reg);
  for (i = n - 1; i >= 0; i--) {
    printf("%.1X", r[i] & 0xf);
  }
  printf("\n");
}

static void
#ifdef __FunctionProto__
set_st(word_64 val)
#else
set_st(val)
word_64 val;
#endif
{
  int i;

  for (i = 0; i < NR_PSTAT; i++) {
    if (val.lo & (1 << i))
      saturn.PSTAT[i] = (unsigned char)1;
    else
      saturn.PSTAT[i] = (unsigned char)0;
  }
}

static void
#ifdef __FunctionProto__
dump_st(void)
#else
dump_st()
#endif
{
  int i;
  unsigned short st = 0;

  for (i = NR_PSTAT - 1; i >= 0; i--) {
    st <<= 1;
    if (saturn.PSTAT[i] != 0) {
      st |= 1;
    }
  }
  printf("    ST:\t%.4X (", st);
  for (i = NR_PSTAT - 1; i > 0; i--) {
    if (saturn.PSTAT[i] != 0) {
      printf("%.1X ", i);
    } else {
      printf("- ");
    }
  }
  if (saturn.PSTAT[0] != 0) {
    printf("%.1X)\n", 0);
  } else {
    printf("-)\n");
  }
}

static void
#ifdef __FunctionProto__
set_hst(word_64 val)
#else
set_hst(val)
word_64 val;
#endif
{
  saturn.XM = 0;
  saturn.SB = 0;
  saturn.SR = 0;
  saturn.MP = 0;
  if (val.lo & 1) saturn.XM = 1;
  if (val.lo & 2) saturn.SB = 1;
  if (val.lo & 4) saturn.SR = 1;
  if (val.lo & 8) saturn.MP = 1;
}

static void
#ifdef __FunctionProto__
dump_hst(void)
#else
dump_hst()
#endif
{
  short hst = 0;
  if (saturn.XM != 0) hst |= 1;
  if (saturn.SB != 0) hst |= 2;
  if (saturn.SR != 0) hst |= 3;
  if (saturn.MP != 0) hst |= 4;
  printf("   HST:\t%.1X    (%s%s%s%s)\n", hst,
         saturn.MP ? "MP ":"-- ", saturn.SR ? "SR ":"-- ",
         saturn.SB ? "SB ":"-- ", saturn.XM ? "XM":"--");
}

static void
#ifdef __FunctionProto__
do_regs(int argc, char **argv)
#else
do_regs(argc, argv)
int argc;
char *argv;
#endif
{
  word_64 val;

  if (argc < 2) {
    /*
     * dump all registers
     */
    printf("CPU is in %s mode. Registers:\n",
           saturn.hexmode == HEX ? "HEX" : "DEC");
    dump_reg("     A", 16, saturn.A);
    dump_reg("     B", 16, saturn.B);
    dump_reg("     C", 16, saturn.C);
    dump_reg("     D", 16, saturn.D);
    printf("    D0:\t%.5lX -> %s\n", saturn.D0, str_nibbles(saturn.D0, 16));
    printf("    D1:\t%.5lX -> %s\n", saturn.D1, str_nibbles(saturn.D1, 16));
    printf("     P:\t%.1X\n", saturn.P);
    Instruction(saturn.PC, instr);
    printf("    PC:\t%.5lX -> %s\n", saturn.PC, instr);
    dump_reg("    R0", 16, saturn.R0);
    dump_reg("    R1", 16, saturn.R1);
    dump_reg("    R2", 16, saturn.R2);
    dump_reg("    R3", 16, saturn.R3);
    dump_reg("    R4", 16, saturn.R4);
    dump_reg("    IN", 4, saturn.IN);
    dump_reg("   OUT", 3, saturn.OUT);
    printf(" CARRY:\t%.1d\n", saturn.CARRY);
    dump_st();
    dump_hst();
  } else if (argc == 2) {
    /*
     * dump specified register
     */
    str_to_upper(argv[1]);
    if (!strcmp("A", argv[1])) {
      dump_reg("     A", 16, saturn.A);
    } else if (!strcmp("B", argv[1])) {
      dump_reg("     B", 16, saturn.B);
    } else if (!strcmp("C", argv[1])) {
      dump_reg("     C", 16, saturn.C);
    } else if (!strcmp("D", argv[1])) {
      dump_reg("     D", 16, saturn.D);
    } else if (!strcmp("D0", argv[1])) {
      printf("    D0:\t%.5lX -> %s\n", saturn.D0, str_nibbles(saturn.D0, 16));
    } else if (!strcmp("D1", argv[1])) {
      printf("    D1:\t%.5lX -> %s\n", saturn.D1, str_nibbles(saturn.D1, 16));
    } else if (!strcmp("P", argv[1])) {
      printf("     P:\t%.1X\n", saturn.P);
    } else if (!strcmp("PC", argv[1])) {
      Instruction(saturn.PC, instr);
      printf("    PC:\t%.5lX -> %s\n", saturn.PC, instr);
    } else if (!strcmp("R0", argv[1])) {
      dump_reg("    R0", 16, saturn.R0);
    } else if (!strcmp("R1", argv[1])) {
      dump_reg("    R1", 16, saturn.R1);
    } else if (!strcmp("R2", argv[1])) {
      dump_reg("    R2", 16, saturn.R2);
    } else if (!strcmp("R3", argv[1])) {
      dump_reg("    R3", 16, saturn.R3);
    } else if (!strcmp("R4", argv[1])) {
      dump_reg("    R4", 16, saturn.R4);
    } else if (!strcmp("IN", argv[1])) {
      dump_reg("    IN", 4, saturn.IN);
    } else if (!strcmp("OUT", argv[1])) {
      dump_reg("   OUT", 3, saturn.OUT);
    } else if (!strcmp("CARRY", argv[1])) {
      printf(" CARRY:\t%.1d\n", saturn.CARRY);
    } else if (!strcmp("CY", argv[1])) {
      printf(" CARRY:\t%.1d\n", saturn.CARRY);
    } else if (!strcmp("ST", argv[1])) {
      dump_st();
    } else if (!strcmp("HST", argv[1])) {
      dump_hst();
    } else {
      printf("No Register %s in CPU.\n", argv[1]);
    }
  } else {
    /*
     * set specified register
     */
    str_to_upper(argv[1]);
    str_to_upper(argv[2]);
    if (decode_64(&val, argv[2])) {
      if (!strcmp("A", argv[1])) {
        set_reg(val, 16, saturn.A);
        dump_reg("     A", 16, saturn.A);
      } else if (!strcmp("B", argv[1])) {
        set_reg(val, 16, saturn.B);
        dump_reg("     B", 16, saturn.B);
      } else if (!strcmp("C", argv[1])) {
        set_reg(val, 16, saturn.C);
        dump_reg("     C", 16, saturn.C);
      } else if (!strcmp("D", argv[1])) {
        set_reg(val, 16, saturn.D);
        dump_reg("     D", 16, saturn.D);
      } else if (!strcmp("D0", argv[1])) {
        saturn.D0 = val.lo & 0xfffff;
        printf("    D0:\t%.5lX -> %s\n", saturn.D0, str_nibbles(saturn.D0, 16));
      } else if (!strcmp("D1", argv[1])) {
        saturn.D1 = val.lo & 0xfffff;
        printf("    D1:\t%.5lX -> %s\n", saturn.D1, str_nibbles(saturn.D1, 16));
      } else if (!strcmp("P", argv[1])) {
        saturn.P = val.lo & 0xf;
        printf("     P:\t%.1X\n", saturn.P);
      } else if (!strcmp("PC", argv[1])) {
        saturn.PC = val.lo & 0xfffff;
        Instruction(saturn.PC, instr);
        printf("    PC:\t%.5lX -> %s\n", saturn.PC, instr);
      } else if (!strcmp("R0", argv[1])) {
        set_reg(val, 16, saturn.R0);
        dump_reg("    R0", 16, saturn.R0);
      } else if (!strcmp("R1", argv[1])) {
        set_reg(val, 16, saturn.R1);
        dump_reg("    R1", 16, saturn.R1);
      } else if (!strcmp("R2", argv[1])) {
        set_reg(val, 16, saturn.R2);
        dump_reg("    R2", 16, saturn.R2);
      } else if (!strcmp("R3", argv[1])) {
        set_reg(val, 16, saturn.R3);
        dump_reg("    R3", 16, saturn.R3);
      } else if (!strcmp("R4", argv[1])) {
        set_reg(val, 16, saturn.R4);
        dump_reg("    R4", 16, saturn.R4);
      } else if (!strcmp("IN", argv[1])) {
        set_reg(val, 4, saturn.IN);
        dump_reg("    IN", 4, saturn.IN);
      } else if (!strcmp("OUT", argv[1])) {
        set_reg(val, 3, saturn.OUT);
        dump_reg("   OUT", 3, saturn.OUT);
      } else if (!strcmp("CARRY", argv[1])) {
        saturn.CARRY = val.lo & 0x1;
        printf(" CARRY:\t%.1d\n", saturn.CARRY);
      } else if (!strcmp("CY", argv[1])) {
        saturn.CARRY = val.lo & 0x1;
        printf(" CARRY:\t%.1d\n", saturn.CARRY);
      } else if (!strcmp("ST", argv[1])) {
        set_st(val);
        dump_st();
      } else if (!strcmp("HST", argv[1])) {
        set_hst(val);
        dump_hst();
      } else {
        printf("No Register %s in CPU.\n", argv[1]);
      }
    }
  }
}

static void
#ifdef __FunctionProto__
do_save(int argc, char **argv)
#else
do_save(argc, argv)
int argc;
char *argv;
#endif
{
  if (write_files()) {
    printf("Saving done.\n");
  } else {
    printf("Saving emulator-state failed.\n");
  }
}

static void
#ifdef __FunctionProto__
do_stat(int argc, char **argv)
#else
do_stat(argc, argv)
int argc;
char *argv;
#endif
{
  printf("Instructions/s: %ld\n", saturn.i_per_s);
  printf("Timer 1 I/TICK: %d\n", saturn.t1_tick);
  printf("Timer 2 I/TICK: %d\n", saturn.t2_tick);
}

static void
#ifdef __FunctionProto__
do_step(int argc, char **argv)
#else
do_step(argc, argv)
int argc;
char *argv;
#endif
{
  step_instruction();
}

static void
#ifdef __FunctionProto__
do_reset(int argc, char **argv)
#else
do_reset(argc, argv)
int argc;
char *argv;
#endif
{
  if (confirm("Do a RESET (PC = 00000)?")) {
    saturn.PC = 0;
    enter_debugger &= ~ILLEGAL_INSTRUCTION;
  }
}

static void
#ifdef __FunctionProto__
do_rstk(int argc, char **argv)
#else
do_rstk(argc, argv)
int argc;
char *argv;
#endif
{
  int i, j;

  Instruction(saturn.PC, instr);
  printf("PC: %.5lX: %s\n", saturn.PC, instr);
  if (saturn.rstkp < 0) {
    printf("Empty return stack.\n");
  } else {
    j = 0;
    for (i = saturn.rstkp; i >= 0; i--) {
      Instruction(saturn.rstk[i], instr);
      printf("%2d: %.5lX: %s\n", j, saturn.rstk[i], instr);
      j++;
    }
  }
}

int
#ifdef __FunctionProto__
debug(void)
#else
debug()
#endif
{
  t1_t2_ticks ticks;
  struct cmd *cmdp;
  char *cp;
  int argc;
  char *argv[MAX_ARGS];
  char *rl;
  static char *cl = (char *)0;
  static char *old_line = (char *)0;
  int i;

  /*
   * update the lcd if necessary
   */
  if (device.display_touched) {
    device.display_touched = 0;
    update_display();
#ifdef USE_SHM
    if (disp.display_update) refresh_display();
#endif
  }
  
  /*
   * debugging is counted as idle time
   */
  stop_timer(RUN_TIMER);
  start_timer(IDLE_TIMER);

  continue_flag = 0;

  if (enter_debugger & ILLEGAL_INSTRUCTION) {
    printf("ILLEGAL INSTRUCTION at %.5lX : %s\n",
           saturn.PC, str_nibbles(saturn.PC, 16));
  }

  do {

    /*
     * print current instruction
     */
    Instruction(saturn.PC, instr);
    printf("%.5lX: %s\n", saturn.PC, instr);

    /*
     * read a command
     */
    rl = readline("x48-debug> ");
    if (rl == (char *)0) {
      continue_flag = 1;
      continue;
    }
    if (*rl == '\0') {
      free(rl);
      if (cl) free(cl);
      if (old_line)
        cl = strcpy((char *)malloc(strlen(old_line)), old_line);
      else
        cl = strcpy((char *)malloc(strlen("(null)")), "(null)");
    } else {
      if (cl) free(cl);
      if (old_line) free(old_line);
      cl = strcpy((char *)malloc(strlen(rl)), rl);
      old_line = strcpy((char *)malloc(strlen(rl)), rl);
      add_history(rl);
      free(rl);
    }

    /*
     * decode the commandline
     */
    cp = strtok(cl, " \t");
    for (cmdp = cmd_tbl; cmdp->name; cmdp++) {
      if (strcmp(cp, cmdp->name) == 0){
        break;
      }
    }

    argc = 0;
    argv[argc++] = cp;
    while ((cp = strtok((char *)0, " \t")) != (char *)0) {
      argv[argc++] = cp;
      if (argc == MAX_ARGS)
        break;
    }
    for (i = argc; i < MAX_ARGS; i++) argv[i] = (char *)NULL;

    /*
     * execute the command, if valid
     */
    if (cmdp->func) {
      (*cmdp->func)(argc, argv);
    } else {
      printf("Undefined command \"%s\". Try \"help\".\n", argv[0]);
    }

  } while (!continue_flag);

  /*
   * adjust the hp48's timers
   */
  ticks = get_t1_t2();
  if (saturn.t2_ctrl & 0x01) {
    saturn.timer2 = set_t2 - ticks.t2_ticks;
  }

  if ((saturn.t2_ctrl & 0x08) == 0 && saturn.timer2 <= 0) {
    if (saturn.t2_ctrl & 0x02) {
      saturn.t2_ctrl |= 0x08;
      do_interupt();
    }
  }

  saturn.timer1 = set_t1 - ticks.t1_ticks;
  if ((saturn.t1_ctrl & 0x08) == 0 && saturn.timer1 <= 0) {
    if (saturn.t1_ctrl & 0x02) {
      saturn.t1_ctrl |= 0x08;
      do_interupt();
    }
  }
  saturn.timer1 &= 0x0f;

  /*
   * restart timers
   */
  stop_timer(IDLE_TIMER);
  start_timer(RUN_TIMER);

  if (enter_debugger & ILLEGAL_INSTRUCTION) {
    printf("Reset (ILLEGAL INSTRUCTION)\n");
    saturn.PC = 0;
  } else {
    printf("Continue.\n");
  }

  enter_debugger = 0;

  return 0;
}

