/*
 *  This file is part of x48, an emulator of the HP-48sx Calculator.
 *  Copyright (C) 1994  Eddie C. Dost  (ecd@dressler.de)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* $Log: init.c,v $
 * Revision 1.8  1994/10/09  20:32:02  ecd
 * deleted extern char lcd_buffer reference.
 *
 * Revision 1.7  1994/10/06  16:30:05  ecd
 * changed char to unsigned
 *
 * Revision 1.6  1994/10/05  08:36:44  ecd
 * changed saturn_config_init()
 *
 * Revision 1.5  1994/09/30  12:37:09  ecd
 * the file ~/.hp48/hp48 now contains a MAGIC and version info, so
 * backward compatibility can be achived
 *
 * Revision 1.4  1994/09/18  15:29:22  ecd
 * turned off unused rcsid message
 *
 * Revision 1.3  1994/09/13  16:57:00  ecd
 * changed to plain X11
 *
 * Revision 1.2  1994/08/31  18:23:21  ecd
 * changed display initialization.
 *
 * Revision 1.1  1994/08/26  11:09:02  ecd
 * Initial revision
 *
 * $Id: init.c,v 1.8 1994/10/09 20:32:02 ecd Exp ecd $
 */


#include "global.h"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/stat.h>
#include <string.h>
#ifdef SUNOS
#include <memory.h>
#endif

#include "hp48.h"
#include "hp48_emu.h"
#include "device.h"
#include "version.h"

typedef struct old_keystate_t {
  int rows[9];
} old_keystate_t;

typedef struct old_saturn_t {
  unsigned char A[16], B[16], C[16], D[16];
  long          d[2];
  int           P;
  long          PC;
  unsigned char R0[16], R1[16], R2[16], R3[16], R4[16];
  unsigned char IN[4];
  unsigned char OUT[3];
  int           CARRY;
  unsigned char PSTAT[NR_PSTAT];
  unsigned char XM, SB, SR, MP;
  unsigned char hexmode;
  long          rstk[NR_RSTK];
  short         rstkp;
  old_keystate_t keybuf;
  unsigned char intenable;
  unsigned char int_pending;
  unsigned char kbd_ien;
  long          configs[NR_CONFIG];
  short         daisy_state;
  long          ram32k;
  long          devices;
  unsigned char disp_io;
  unsigned char contrast_ctrl;
  unsigned char disp_test;
  unsigned int  crc;
  unsigned char power_status;
  unsigned char power_ctrl;
  unsigned char mode;
  unsigned char annunc;
  unsigned char baud;
  unsigned char card_ctrl;
  unsigned char card_status;
  unsigned char io_ctrl;
  unsigned char rcs;
  unsigned char tcs;
  unsigned char rbr;
  unsigned char tbr;
  unsigned char sreq;
  unsigned char ir_ctrl;
  unsigned char base_off;
  unsigned char lcr;
  unsigned char lbr;
  unsigned char scratch;
  unsigned char base_nibble;
  long          disp_addr;
  long          line_offset;
  long          line_count;
  long          unknown;
  unsigned char t1_ctrl;
  unsigned char t2_ctrl;
  long          menu_addr;
  long          unknown2;
  int           timer1;
  long          timer2;
  long          t1_instr;
  long          t2_instr;
  unsigned char *rom;
  unsigned char *ram;
  unsigned char *port1;
  unsigned char *port2;
} old_saturn_t;

old_saturn_t old_saturn;

void
#ifdef __FunctionProto__
saturn_config_init(void)
#else
saturn_config_init()
#endif
{
  saturn.version[0] = VERSION_MAJOR;
  saturn.version[1] = VERSION_MINOR;
  saturn.version[2] = PATCHLEVEL;
  saturn.version[3] = 0;
  memset(&device, 0, sizeof(device));
  device.display_touched = 1;
  device.ann_touched = 1;
  write_nibble(0x10f, 0xf);     /* CARDSTAT */
  write_nibble(0x111, 0x0);     /* RCS */
  write_nibble(0x112, 0x0);     /* TCS */
  write_nibble(0x11d, 0x0);     /* LBR */
}

int
#ifdef __FunctionProto__
read_rom(const char *fname)
#else
read_rom(fname)
const char *fname;
#endif
{
  FILE *in;

  if ((in = fopen(fname, "r")) == NULL) {
    return 0;
  }
  if ((saturn.rom = (unsigned char *)malloc(ROM_SIZE)) == NULL) {
    fclose(in);
    return 0;
  }
  if (fread(saturn.rom, 1, ROM_SIZE, in) != ROM_SIZE) {
    fclose(in);
    free(saturn.rom);
    return 0;
  }
  fclose(in);

  if ((saturn.ram = (unsigned char *)malloc(RAM_SIZE)) == NULL) {
    fprintf(stderr, "Can\'t malloc RAM\n");
    return 0;
  }
  memset(saturn.ram, 0, RAM_SIZE);

  if ((saturn.port1 = (unsigned char *)malloc(PORT1_SIZE)) == NULL) {
    fprintf(stderr, "Can\'t malloc PORT1\n");
    return 0;
  }
  memset(saturn.port1, 0, PORT1_SIZE);

  if ((saturn.port2 = (unsigned char *)malloc(PORT2_SIZE)) == NULL) {
    fprintf(stderr, "Can\'t malloc PORT2\n");
    return 0;
  }
  memset(saturn.port2, 0, PORT2_SIZE);

  return 1;
}

void
#ifdef __FunctionProto__
init_saturn(void)
#else
init_saturn()
#endif
{
  memset(&saturn, 0, sizeof(saturn));
  saturn.ram32k = 0x70000;
  saturn.devices = 0x00100;
  saturn.PC = 0x00000;
  saturn.magic = X48_MAGIC;
  saturn.t1_tick = 8192;
  saturn.t2_tick = 16;
  saturn.i_per_s = 0;
  saturn.version[0] = VERSION_MAJOR;
  saturn.version[1] = VERSION_MINOR;
  saturn.version[2] = PATCHLEVEL;
  saturn.version[3] = '\0';
  saturn.hexmode = HEX;
  saturn.rstkp = -1;
  saturn.intenable = 1;
  saturn.int_pending = 0;
  saturn.kbd_ien = 1;
  dev_memory_init();
}

void
#ifdef __FunctionProto__
copy_old_saturn(old_saturn_t *old, saturn_t *new)
#else
copy_old_saturn(old, new)
old_saturn_t *old;
saturn_t *new;
#endif
{
  int i;

  memcpy(&(new->A[0]), &(old->A[0]), 16);
  memcpy(&(new->B[0]), &(old->B[0]), 16);
  memcpy(&(new->C[0]), &(old->C[0]), 16);
  memcpy(&(new->D[0]), &(old->D[0]), 16);
  new->d[0] = old->d[0];
  new->d[1] = old->d[1];
  new->P = old->P;
  new->PC = old->PC;
  memcpy(&(new->R0[0]), &(old->R0[0]), 16);
  memcpy(&(new->R1[0]), &(old->R1[0]), 16);
  memcpy(&(new->R2[0]), &(old->R2[0]), 16);
  memcpy(&(new->R3[0]), &(old->R3[0]), 16);
  memcpy(&(new->R4[0]), &(old->R4[0]), 16);
  memcpy(&(new->IN[0]), &(old->IN[0]), 4);
  memcpy(&(new->OUT[0]), &(old->OUT[0]), 3);
  new->CARRY = old->CARRY;
  memcpy(&(new->PSTAT[0]), &(old->PSTAT[0]), NR_PSTAT);
  new->XM = old->XM;
  new->SB = old->SB;
  new->SR = old->SR;
  new->MP = old->MP;
  new->hexmode = old->hexmode;
  memcpy(&(new->rstk[0]), &(old->rstk[0]), NR_RSTK * sizeof(word_20));
  new->rstkp = old->rstkp;
  for (i = 0; i < 9; i++) {
    new->keybuf.rows[i] = old->keybuf.rows[i];
  }
  new->intenable = old->intenable;
  new->int_pending = old->int_pending;
  new->kbd_ien = old->kbd_ien;
  memcpy(&(new->configs[0]), &(old->configs[0]), NR_CONFIG * sizeof(word_20));
  new->daisy_state = old->daisy_state;
  new->ram32k = old->ram32k;
  new->devices = old->devices;
  new->disp_io = old->disp_io;
  new->contrast_ctrl = old->contrast_ctrl;
  new->disp_test = old->disp_test;
  new->crc = old->crc;
  new->power_status = old->power_status;
  new->power_ctrl = old->power_ctrl;
  new->mode = old->mode;
  new->annunc = old->annunc;
  new->baud = old->baud;
  new->card_ctrl = old->card_ctrl;
  new->card_status = old->card_status;
  new->io_ctrl = old->io_ctrl;
  new->rcs = old->rcs;
  new->tcs = old->tcs;
  new->rbr = old->rbr;
  new->tbr = old->tbr;
  new->sreq = old->sreq;
  new->ir_ctrl = old->ir_ctrl;
  new->base_off = old->base_off;
  new->lcr = old->lcr;
  new->lbr = old->lbr;
  new->scratch = old->scratch;
  new->base_nibble = old->base_nibble;
  new->disp_addr = old->disp_addr;
  new->line_offset = old->line_offset;
  new->line_count = old->line_count;
  new->unknown = old->unknown;
  new->t1_ctrl = old->t1_ctrl;
  new->t2_ctrl = old->t2_ctrl;
  new->menu_addr = old->menu_addr;
  new->unknown2 = old->unknown2;
  new->timer1 = old->timer1;
  new->timer2 = old->timer2;
  new->t1_instr = old->t1_instr;
  new->t2_instr = old->t2_instr;
}

int
#ifdef __FunctionProto__
read_8(FILE *fp, word_8 *var)
#else
read_8(fp, var)
FILE *fp;
word_8 *var;
#endif
{
  unsigned char tmp;

  if (fread(&tmp, 1, 1, fp) != 1) {
    fprintf(stderr, "can\'t read word_8\n");
    return 0;
  }
  *var = tmp;
  return 1;
}

int
#ifdef __FunctionProto__
read_char(FILE *fp, char *var)
#else
read_char(fp, var)
FILE *fp;
char *var;
#endif
{
  char tmp;

  if (fread(&tmp, 1, 1, fp) != 1) {
    fprintf(stderr, "can\'t read char\n");
    return 0;
  }
  *var = tmp;
  return 1;
}

int
#ifdef __FunctionProto__
read_16(FILE *fp, word_16 *var)
#else
read_16(fp, var)
FILE *fp;
word_16 *var;
#endif
{
  unsigned char tmp[2];

  if (fread(&tmp[0], 1, 2, fp) != 2) {
    fprintf(stderr, "can\'t read word_16\n");
    return 0;
  }
  *var = tmp[0] << 8;
  *var |= tmp[1];
  return 1;
}

int
#ifdef __FunctionProto__
read_32(FILE *fp, word_32 *var)
#else
read_32(fp, var)
FILE *fp;
word_32 *var;
#endif
{
  unsigned char tmp[4];

  if (fread(&tmp[0], 1, 4, fp) != 4) {
    fprintf(stderr, "can\'t read word_32\n");
    return 0;
  }
  *var = tmp[0] << 24;
  *var |= tmp[1] << 16;
  *var |= tmp[2] << 8;
  *var |= tmp[3];
  return 1;
}

int
#ifdef __FunctionProto__
read_u_long(FILE *fp, unsigned long *var)
#else
read_u_long(fp, var)
FILE *fp;
unsigned long*var;
#endif
{
  unsigned char tmp[4];

  if (fread(&tmp[0], 1, 4, fp) != 4) {
    fprintf(stderr, "can\'t read unsigned long\n");
    return 0;
  }
  *var = tmp[0] << 24;
  *var |= tmp[1] << 16;
  *var |= tmp[2] << 8;
  *var |= tmp[3];
  return 1;
}

int
#ifdef __FunctionProto__
read_version_0_3_0_file(FILE *fp)
#else
read_version_0_3_0_file(fp)
FILE *fp;
#endif
{
  int i;

  /*
   * version 0.3.x, read in the saturn_t struct
   */
  for (i = 0; i < 16; i++)
    if(!read_8(fp, &saturn.A[i]))
      return 0;
  for (i = 0; i < 16; i++)
    if (!read_8(fp, &saturn.B[i]))
      return 0;
  for (i = 0; i < 16; i++)
    if (!read_8(fp, &saturn.C[i]))
      return 0;
  for (i = 0; i < 16; i++)
    if (!read_8(fp, &saturn.D[i]))
      return 0;
  if (!read_32(fp, &saturn.d[0])) return 0;
  if (!read_32(fp, &saturn.d[1])) return 0;
  if (!read_8(fp, &saturn.P)) return 0;
  if (!read_32(fp, &saturn.PC)) return 0;
  for (i = 0; i < 16; i++)
    if (!read_8(fp, &saturn.R0[i]))
      return 0;
  for (i = 0; i < 16; i++)
    if (!read_8(fp, &saturn.R1[i]))
      return 0;
  for (i = 0; i < 16; i++)
    if (!read_8(fp, &saturn.R2[i]))
      return 0;
  for (i = 0; i < 16; i++)
    if (!read_8(fp, &saturn.R3[i]))
      return 0;
  for (i = 0; i < 16; i++)
    if (!read_8(fp, &saturn.R4[i]))
      return 0;
  for (i = 0; i < 4; i++)
    if (!read_8(fp, &saturn.IN[i]))
      return 0;
  for (i = 0; i < 3; i++)
    if (!read_8(fp, &saturn.OUT[i]))
      return 0;
  if (!read_8(fp, &saturn.CARRY))
     return 0;
  for (i = 0; i < NR_PSTAT; i++)
    if (!read_8(fp, &saturn.PSTAT[i]))
      return 0;
  if (!read_8(fp, &saturn.XM)) return 0;
  if (!read_8(fp, &saturn.SB)) return 0;
  if (!read_8(fp, &saturn.SR)) return 0;
  if (!read_8(fp, &saturn.MP)) return 0;
  if (!read_8(fp, &saturn.hexmode)) return 0;
  for (i = 0; i < NR_RSTK; i++)
    if (!read_32(fp, &saturn.rstk[i]))
      return 0;
  if (!read_16(fp, &saturn.rstkp)) return 0;
  for (i = 0; i < 9; i++)
    if (!read_16(fp, &saturn.keybuf.rows[i]))
      return 0;
  if (!read_8(fp, &saturn.intenable)) return 0;
  if (!read_8(fp, &saturn.int_pending)) return 0;
  if (!read_8(fp, &saturn.kbd_ien)) return 0;
  for (i = 0; i < NR_CONFIG; i++)
    if (!read_32(fp, &saturn.configs[i]))
      return 0;
  if (!read_16(fp, &saturn.daisy_state)) return 0;
  if (!read_32(fp, &saturn.ram32k)) return 0;
  if (!read_32(fp, &saturn.devices)) return 0;
  if (!read_8(fp, &saturn.disp_io)) return 0;
  if (!read_8(fp, &saturn.contrast_ctrl)) return 0;
  if (!read_8(fp, &saturn.disp_test)) return 0;
  if (!read_16(fp, &saturn.crc)) return 0;
  if (!read_8(fp, &saturn.power_status)) return 0;
  if (!read_8(fp, &saturn.power_ctrl)) return 0;
  if (!read_8(fp, &saturn.mode)) return 0;
  if (!read_8(fp, &saturn.annunc)) return 0;
  if (!read_8(fp, &saturn.baud)) return 0;
  if (!read_8(fp, &saturn.card_ctrl)) return 0;
  if (!read_8(fp, &saturn.card_status)) return 0;
  if (!read_8(fp, &saturn.io_ctrl)) return 0;
  if (!read_8(fp, &saturn.rcs)) return 0;
  if (!read_8(fp, &saturn.tcs)) return 0;
  if (!read_8(fp, &saturn.rbr)) return 0;
  if (!read_8(fp, &saturn.tbr)) return 0;
  if (!read_8(fp, &saturn.sreq)) return 0;
  if (!read_8(fp, &saturn.ir_ctrl)) return 0;
  if (!read_8(fp, &saturn.base_off)) return 0;
  if (!read_8(fp, &saturn.lcr)) return 0;
  if (!read_8(fp, &saturn.lbr)) return 0;
  if (!read_8(fp, &saturn.scratch)) return 0;
  if (!read_8(fp, &saturn.base_nibble)) return 0;
  if (!read_32(fp, &saturn.disp_addr)) return 0;
  if (!read_16(fp, &saturn.line_offset)) return 0;
  if (!read_8(fp, &saturn.line_count)) return 0;
  if (!read_16(fp, &saturn.unknown)) return 0;
  if (!read_8(fp, &saturn.t1_ctrl)) return 0;
  if (!read_8(fp, &saturn.t2_ctrl)) return 0;
  if (!read_32(fp, &saturn.menu_addr)) return 0;
  if (!read_8(fp, &saturn.unknown2)) return 0;
  if (!read_char(fp, &saturn.timer1)) return 0;
  if (!read_32(fp, &saturn.timer2)) return 0;
  if (!read_32(fp, &saturn.t1_instr)) return 0;
  if (!read_32(fp, &saturn.t2_instr)) return 0;
  if (!read_16(fp, &saturn.t1_tick)) return 0;
  if (!read_16(fp, &saturn.t2_tick)) return 0;
  if (!read_32(fp, &saturn.i_per_s)) return 0;
  return 1;
}

int
#ifdef __FunctionProto__
read_files(void)
#else
read_files()
#endif
{
  char *home;
  char path[256];
  char fnam[256];
  unsigned long v1, v2;
  int i, read_version;
  FILE *fp;

  if ((home = getenv("HOME")) == NULL) {
    fprintf(stderr, "can\'t getenv $HOME\n");
    return 0;
  }
  strcpy(path, home);
  strcat(path, "/.hp48/");

  strcpy(fnam, path);
  strcat(fnam, "hp48");
  if ((fp = fopen(fnam, "r")) == NULL) {
    fprintf(stderr, "can\'t open %s\n", fnam);
    return 0;
  }

  /*
   * ok, file is open, try to read the MAGIC number
   */
  read_u_long(fp, &saturn.magic);

  if (saturn.magic != X48_MAGIC) {
    /*
     * no MAGIC number, try to read old format file
     */
    fseek(fp, 0, SEEK_SET);
    if (fread(&old_saturn, 1, sizeof(old_saturn), fp) == sizeof(old_saturn)) {
      /*
       * seems to work
       */
      copy_old_saturn(&old_saturn, &saturn);
      fprintf(stderr, "%s seems to be an old version file\n", fnam);
      saturn.magic = X48_MAGIC;
      saturn.t1_tick = 8192;
      saturn.t2_tick = 16;
      saturn.i_per_s = 0;
      saturn.version[0] = VERSION_MAJOR;
      saturn.version[1] = VERSION_MINOR;
      saturn.version[2] = PATCHLEVEL;
      saturn.version[3] = '\0';
    } else {
      /*
       * no, initialize
       */
      fprintf(stderr, "can\'t handle %s\n", fnam);
      init_saturn();
    }
  } else {
    /*
     * MAGIC ok, read and compare the version
     */
    read_version = 1;
    for (i = 0; i < 4; i++) {
      if (!read_char(fp, &saturn.version[i])) {
        fprintf(stderr, "can\'t read Version\n");
        read_version = 0;
      }
    }
    
    if (read_version) {
      v1 = saturn.version[0] << 24;
      v1 |= saturn.version[1] << 16;
      v1 |= saturn.version[2] << 8;
      v1 |= saturn.version[3];
      v2 = VERSION_MAJOR << 24;
      v2 |= VERSION_MINOR << 16;
      v2 |= PATCHLEVEL << 8;

      if (v1 < v2) {
        fprintf(stderr, "%s: version %d.%d.%d file, converting\n", fnam,
                saturn.version[0], saturn.version[1], saturn.version[2]);
      } else if (v2 < v1) {
        fprintf(stderr, "%s: version %d.%d.%d file, trying ...\n", fnam,
                saturn.version[0], saturn.version[1], saturn.version[2]);
      }
  
      if (v1 <= FILE_VERSION) {
        /*
         * read a version 0.3.x file
         */
        if (!read_version_0_3_0_file(fp)) {
          fprintf(stderr, "can\'t handle %s\n", fnam);
          init_saturn();
        }
      } else if (v2 > v1) {
        /*
         * try to read latest version file
         */
        if (!read_version_0_3_0_file(fp)) {
          fprintf(stderr, "can\'t handle %s\n", fnam);
          init_saturn();
        }
      } else {
        /*
         * should never get here
         */
        init_saturn();
      }
    }
  }
  fclose(fp);

  if (!(saturn.ram32k == 0x70000 || saturn.ram32k == 0xf0000 ||
      saturn.ram32k == 0xfc000 || saturn.ram32k == 0xfe000)) {
    fprintf(stderr, "Illegal RAM configuration: 0x%lx\n", saturn.ram32k);
    init_saturn();
  }
  if (!(saturn.devices == 0x00000 || saturn.devices == 0x00100)) {
    fprintf(stderr, "Illegal DEVICE configuration: 0x%lx\n", saturn.devices);
    init_saturn();
  }
  if (saturn.P > 0xf) {
    fprintf(stderr, "P out of range: 0x%x\n", saturn.P);
    init_saturn();
  }
  if (!(saturn.hexmode == DEC || saturn.hexmode == HEX)) {
    fprintf(stderr, "Illegal hexmode setting: 0x%x\n", saturn.hexmode);
    init_saturn();
  }
  if (saturn.t1_tick == 0 || saturn.t2_tick == 0) {
    fprintf(stderr, "Illegal timer speed: T1 %d, T2 %d\n",
            saturn.t1_tick, saturn.t2_tick);
    init_saturn();
  }

  saturn.rom = (unsigned char *)NULL;
  if ((saturn.rom = (unsigned char *)malloc(ROM_SIZE)) == NULL) {
    fprintf(stderr, "can\'t malloc rom[%d]\n", ROM_SIZE);
    exit (1);
  }
  strcpy(fnam, path);
  strcat(fnam, "rom");
  if ((fp = fopen(fnam, "r")) == NULL) {
    fprintf(stderr, "can\'t open %s\n", fnam);
    return 0;
  }
  if (fread(saturn.rom, 1, ROM_SIZE, fp) != ROM_SIZE) {
    fprintf(stderr, "can\'t read %s\n", fnam);
    fclose(fp);
    return 0;
  }
  fclose(fp);

  saturn_config_init();

  saturn.ram = (unsigned char *)NULL;
  if ((saturn.ram = (unsigned char *)malloc(RAM_SIZE)) == NULL) {
    fprintf(stderr, "can\'t malloc ram[%d]\n", RAM_SIZE);
    exit (1);
  }
  strcpy(fnam, path);
  strcat(fnam, "ram");
  if ((fp = fopen(fnam, "r")) == NULL) {
    fprintf(stderr, "can\'t open %s\n", fnam);
    return 0;
  }
  if (fread(saturn.ram, 1, RAM_SIZE, fp) != RAM_SIZE) {
    fprintf(stderr, "can\'t read %s\n", fnam);
    fclose(fp);
  }
  fclose(fp);

  saturn.port1 = (unsigned char *)NULL;
  if ((saturn.port1 = (unsigned char *)malloc(PORT1_SIZE)) == NULL) {
    fprintf(stderr, "can\'t malloc port1[%d]\n", PORT1_SIZE);
    exit (1);
  }
  strcpy(fnam, path);
  strcat(fnam, "port1");
  if ((fp = fopen(fnam, "r")) == NULL) {
    fprintf(stderr, "can\'t open %s\n", fnam);
    return 0;
  }
  if (fread(saturn.port1, 1, PORT1_SIZE, fp) != PORT1_SIZE) {
    fprintf(stderr, "can\'t read %s\n", fnam);
    fclose(fp);
  }
  fclose(fp);

  saturn.port2 = (unsigned char *)NULL;
  if ((saturn.port2 = (unsigned char *)malloc(PORT2_SIZE)) == NULL) {
    fprintf(stderr, "can\'t malloc port2[%d]\n", PORT2_SIZE);
    exit (1);
  }
  strcpy(fnam, path);
  strcat(fnam, "port2");
  if ((fp = fopen(fnam, "r")) == NULL) {
    fprintf(stderr, "can\'t open %s\n", fnam);
    return 0;
  }
  if (fread(saturn.port2, 1, PORT2_SIZE, fp) != PORT2_SIZE) {
    fprintf(stderr, "can\'t read %s\n", fnam);
    fclose(fp);
  }
  fclose(fp);

  return 1;
}

int
#ifdef __FunctionProto__
write_8(FILE *fp, word_8 *var)
#else
write_8(fp, var)
FILE *fp;
word_8 *var;
#endif
{
  unsigned char tmp;

  tmp = *var;
  if (fwrite(&tmp, 1, 1, fp) != 1) {
    fprintf(stderr, "can\'t write word_8\n");
    return 0;
  }
  return 1;
}

int
#ifdef __FunctionProto__
write_char(FILE *fp, char *var)
#else
write_char(fp, var)
FILE *fp;
char *var;
#endif
{
  char tmp;

  tmp = *var;
  if (fwrite(&tmp, 1, 1, fp) != 1) {
    fprintf(stderr, "can\'t write char\n");
    return 0;
  }
  return 1;
}

int
#ifdef __FunctionProto__
write_16(FILE *fp, word_16 *var)
#else
write_16(fp, var)
FILE *fp;
word_16 *var;
#endif
{
  unsigned char tmp[2];

  tmp[0] = (*var >> 8) & 0xff;
  tmp[1] = *var & 0xff;
  if (fwrite(&tmp[0], 1, 2, fp) != 2) {
    fprintf(stderr, "can\'t write word_16\n");
    return 0;
  }
  return 1;
}

int
#ifdef __FunctionProto__
write_32(FILE *fp, word_32 *var)
#else
write_32(fp, var)
FILE *fp;
word_32 *var;
#endif
{
  unsigned char tmp[4];

  tmp[0] = (*var >> 24) & 0xff;
  tmp[1] = (*var >> 16) & 0xff;
  tmp[2] = (*var >> 8) & 0xff;
  tmp[3] = *var & 0xff;
  if (fwrite(&tmp[0], 1, 4, fp) != 4) {
    fprintf(stderr, "can\'t write word_32\n");
    return 0;
  }
  return 1;
}

int
#ifdef __FunctionProto__
write_u_long(FILE *fp, unsigned long *var)
#else
write_u_long(fp, var)
FILE *fp;
unsigned long*var;
#endif
{
  unsigned char tmp[4];

  tmp[0] = (*var >> 24) & 0xff;
  tmp[1] = (*var >> 16) & 0xff;
  tmp[2] = (*var >> 8) & 0xff;
  tmp[3] = *var & 0xff;
  if (fwrite(&tmp[0], 1, 4, fp) != 4) {
    fprintf(stderr, "can\'t write unsigned long\n");
    return 0;
  }
  return 1;
}

int
#ifdef __FunctionProto__
write_files(void)
#else
write_files()
#endif
{
  char *home;
  char path[256];
  char fnam[256];
  struct stat st;
  int i, make_dir;
  FILE *fp;

  make_dir = 0;
  if ((home = getenv("HOME")) != NULL) {
    strcpy(path, home);
    strcat(path, "/.hp48");
    if (stat(path, &st) == -1) {
      if (errno == ENOENT) {
        make_dir = 1;
      } else {
        fprintf(stderr, "can\'t stat %s, saving to /tmp\n", path);
        strcpy(path, "/tmp");
      }
    } else {
      if (!S_ISDIR(st.st_mode)) {
        fprintf(stderr, "%s is no directory, saving to /tmp\n", path);
        strcpy(path, "/tmp");
      }
    }
  } else {
    strcpy(path, "/tmp");
  }
  if (make_dir) {
    if (mkdir(path, 0777) == -1) {
      fprintf(stderr, "can\'t mkdir %s, saving to /tmp\n", path);
      strcpy(path, "/tmp");
    }
  }
  strcat(path, "/");
  strcpy(fnam, path);
  strcat(fnam, "hp48");
  if ((fp = fopen(fnam, "w")) == NULL) {
    fprintf(stderr, "can\'t open %s, no saving done\n", fnam);
    return 0;
  }

  /*
   * write the hp48 config file
   */
  write_32(fp, &saturn.magic);
  for (i = 0; i < 4; i++) write_char(fp, &saturn.version[i]);
  for (i = 0; i < 16; i++) write_8(fp, &saturn.A[i]);
  for (i = 0; i < 16; i++) write_8(fp, &saturn.B[i]);
  for (i = 0; i < 16; i++) write_8(fp, &saturn.C[i]);
  for (i = 0; i < 16; i++) write_8(fp, &saturn.D[i]);
  write_32(fp, &saturn.d[0]);
  write_32(fp, &saturn.d[1]);
  write_8(fp, &saturn.P);
  write_32(fp, &saturn.PC);
  for (i = 0; i < 16; i++) write_8(fp, &saturn.R0[i]);
  for (i = 0; i < 16; i++) write_8(fp, &saturn.R1[i]);
  for (i = 0; i < 16; i++) write_8(fp, &saturn.R2[i]);
  for (i = 0; i < 16; i++) write_8(fp, &saturn.R3[i]);
  for (i = 0; i < 16; i++) write_8(fp, &saturn.R4[i]);
  for (i = 0; i < 4; i++) write_8(fp, &saturn.IN[i]);
  for (i = 0; i < 3; i++) write_8(fp, &saturn.OUT[i]);
  write_8(fp, &saturn.CARRY);
  for (i = 0; i < NR_PSTAT; i++) write_8(fp, &saturn.PSTAT[i]);
  write_8(fp, &saturn.XM);
  write_8(fp, &saturn.SB);
  write_8(fp, &saturn.SR);
  write_8(fp, &saturn.MP);
  write_8(fp, &saturn.hexmode);
  for (i = 0; i < NR_RSTK; i++) write_32(fp, &saturn.rstk[i]);
  write_16(fp, &saturn.rstkp);
  for (i = 0; i < 9; i++) write_16(fp, &saturn.keybuf.rows[i]);
  write_8(fp, &saturn.intenable);
  write_8(fp, &saturn.int_pending);
  write_8(fp, &saturn.kbd_ien);
  for (i = 0; i < NR_CONFIG; i++) write_32(fp, &saturn.configs[i]);
  write_16(fp, &saturn.daisy_state);
  write_32(fp, &saturn.ram32k);
  write_32(fp, &saturn.devices);
  write_8(fp, &saturn.disp_io);
  write_8(fp, &saturn.contrast_ctrl);
  write_8(fp, &saturn.disp_test);
  write_16(fp, &saturn.crc);
  write_8(fp, &saturn.power_status);
  write_8(fp, &saturn.power_ctrl);
  write_8(fp, &saturn.mode);
  write_8(fp, &saturn.annunc);
  write_8(fp, &saturn.baud);
  write_8(fp, &saturn.card_ctrl);
  write_8(fp, &saturn.card_status);
  write_8(fp, &saturn.io_ctrl);
  write_8(fp, &saturn.rcs);
  write_8(fp, &saturn.tcs);
  write_8(fp, &saturn.rbr);
  write_8(fp, &saturn.tbr);
  write_8(fp, &saturn.sreq);
  write_8(fp, &saturn.ir_ctrl);
  write_8(fp, &saturn.base_off);
  write_8(fp, &saturn.lcr);
  write_8(fp, &saturn.lbr);
  write_8(fp, &saturn.scratch);
  write_8(fp, &saturn.base_nibble);
  write_32(fp, &saturn.disp_addr);
  write_16(fp, &saturn.line_offset);
  write_8(fp, &saturn.line_count);
  write_16(fp, &saturn.unknown);
  write_8(fp, &saturn.t1_ctrl);
  write_8(fp, &saturn.t2_ctrl);
  write_32(fp, &saturn.menu_addr);
  write_8(fp, &saturn.unknown2);
  write_char(fp, &saturn.timer1);
  write_32(fp, &saturn.timer2);
  write_32(fp, &saturn.t1_instr);
  write_32(fp, &saturn.t2_instr);
  write_16(fp, &saturn.t1_tick);
  write_16(fp, &saturn.t2_tick);
  write_32(fp, &saturn.i_per_s);
  fclose(fp);

  strcpy(fnam, path);
  strcat(fnam, "rom");
  if ((fp = fopen(fnam, "w")) == NULL) {
    fprintf(stderr, "can\'t open %s, no saving done\n", fnam);
    return 0;
  }
  if (fwrite(saturn.rom, 1, ROM_SIZE, fp) != ROM_SIZE) {
    fprintf(stderr, "can\'t write to %s, no saving done\n", fnam);
    fclose(fp);
    return 0;
  }
  fclose(fp);

  strcpy(fnam, path);
  strcat(fnam, "ram");
  if ((fp = fopen(fnam, "w")) == NULL) {
    fprintf(stderr, "can\'t open %s, no saving done\n", fnam);
    return 0;
  }
  if (fwrite(saturn.ram, 1, RAM_SIZE, fp) != RAM_SIZE) {
    fprintf(stderr, "can\'t write to %s, no saving done\n", fnam);
    fclose(fp);
    return 0;
  }
  fclose(fp);

  strcpy(fnam, path);
  strcat(fnam, "port1");
  if ((fp = fopen(fnam, "w")) == NULL) {
    fprintf(stderr, "can\'t open %s, no saving done\n", fnam);
    return 0;
  }
  if (fwrite(saturn.port1, 1, PORT1_SIZE, fp) != PORT1_SIZE) {
    fprintf(stderr, "can\'t write to %s, no saving done\n", fnam);
    fclose(fp);
    return 0;
  }
  fclose(fp);

  strcpy(fnam, path);
  strcat(fnam, "port2");
  if ((fp = fopen(fnam, "w")) == NULL) {
    fprintf(stderr, "can\'t open %s, no saving done\n", fnam);
    return 0;
  }
  if (fwrite(saturn.port2, 1, PORT2_SIZE, fp) != PORT2_SIZE) {
    fprintf(stderr, "can\'t write to %s, no saving done\n", fnam);
    fclose(fp);
    return 0;
  }
  fclose(fp);

  return 1;
}

int 
#ifdef __FunctionProto__
init_emulator(void)
#else
init_emulator()
#endif
{
  serial_init();
  if (!read_files()) {
    init_saturn();
    if (!read_rom(CORE_FILE)) {
      fprintf(stderr, "%s: can\'t read corefile %s\n", progname, CORE_FILE);
      exit(1);
    }
  }
  
  return 0;
}

int
#ifdef __FunctionProto__
exit_emulator(void)
#else
exit_emulator()
#endif
{
  write_files();
  return 1;
}

