/*
 *  This file is part of x48, an emulator of the HP-48sx Calculator.
 *  Copyright (C) 1994  Eddie C. Dost  (ecd@dressler.de)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* $Log: memory.c,v $
 * Revision 1.9  1994/10/09  20:32:02  ecd
 * added support for bit offset, changed handling of display_t
 * variables.
 *
 * Revision 1.8  1994/10/06  16:30:05  ecd
 * changed char to unsigned
 *
 * Revision 1.7  1994/10/05  08:36:44  ecd
 * removed display addr queue
 *
 * Revision 1.6  1994/10/01  10:12:53  ecd
 * added schedule_event = 0 for any display touch
 *
 * Revision 1.5  1994/09/30  12:37:09  ecd
 * support for faster display and scheduler
 *
 * Revision 1.4  1994/09/18  15:29:22  ecd
 * turned off unused rcsid message
 *
 * Revision 1.3  1994/09/13  16:57:00  ecd
 * changed to plain X11
 *
 * Revision 1.2  1994/08/31  18:23:21  ecd
 * changed memory access routines.
 *
 * Revision 1.1  1994/08/26  11:09:02  ecd
 * Initial revision
 *
 * $Id: memory.c,v 1.9 1994/10/09 20:32:02 ecd Exp ecd $
 */

#include "global.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#ifdef SUNOS
#include <memory.h>
#endif

#include "hp48.h"
#include "device.h"
#include "hp48_emu.h"
#include "x48_x11.h"

extern int device_check;

/* #define DEBUG_SERIAL */
/* #define DEBUG_DISPLAY */
/* #define DEBUG_IR */

long nibble_masks[16] = {
  0x0000000f,
  0x000000f0,
  0x00000f00,
  0x0000f000,
  0x000f0000,
  0x00f00000,
  0x0f000000,
  0xf0000000,
  0x0000000f,
  0x000000f0,
  0x00000f00,
  0x0000f000,
  0x000f0000,
  0x00f00000,
  0x0f000000,
  0xf0000000
};

static int line_counter = -1;

static inline int
#ifdef __FunctionProto__
calc_crc(int nib)
#else
calc_crc(nib)
int nib;
#endif
{
  saturn.crc = (saturn.crc >> 4) ^ (((saturn.crc ^ nib) & 0xf) * 0x1081);
  return nib;
}

void
#ifdef __FunctionProto__
write_dev_mem(long addr, int val)
#else
write_dev_mem(addr, val)
long addr;
int val;
#endif
{
  static int old_line_offset = -1;

  device_check = 1;
  schedule_event = 0;
  switch (addr) {
    case 0x100:					/* DISPIO */
      if (val != saturn.disp_io) {
        saturn.disp_io = val;
        display.on = (val & 0x8) >> 3;
        display.offset = val & 0x7;
        disp.offset = 2 * display.offset;
        if (display.offset > 3)
          display.nibs_per_line =
                  (NIBBLES_PER_ROW+saturn.line_offset+2) & 0xfff;
        else
          display.nibs_per_line =
                  (NIBBLES_PER_ROW+saturn.line_offset) & 0xfff;
        display.disp_end = display.disp_start +
		           (display.nibs_per_line * (display.lines + 1));
        device.display_touched = DISP_INSTR_OFF;
      }
#ifdef DEBUG_DISPLAY
      fprintf(stderr, "%.5lx: DISP ON: %x\n", saturn.PC, display.on);
#endif
      return;
    case 0x101:						/* CONTRAST CONTROL */
      saturn.contrast_ctrl = val;
      display.contrast = val;
      device.contrast_touched = 1;
      return;
    case 0x102: case 0x103:				/* DISPLAY TEST */
      saturn.disp_test &= ~nibble_masks[addr - 0x102];
      saturn.disp_test |= val << ((addr - 0x102) * 4);
#ifdef DEBUG_DISPLAY
      fprintf(stderr, "%.5lx: DISP TEST: %x\n", saturn.PC, saturn.disp_test);
#endif
      device.disp_test_touched = 1;
      return;
    case 0x104: case 0x105: case 0x106: case 0x107:	/* CRC */
      saturn.crc &= ~nibble_masks[addr - 0x104];
      saturn.crc |= val << ((addr - 0x104) * 4);
      return;
    case 0x108:						/* POWER STATUS */
      saturn.power_status = val;
      device.power_status_touched = 1;
      return;
    case 0x109:						/* POWER CONTROL */
      saturn.power_ctrl = val;
      device.power_ctrl_touched = 1;
      return;
    case 0x10a:						/* MODE */
      saturn.mode = val;
      device.mode_touched = 1;
      return;
    case 0x10b: case 0x10c:				/* ANNUNC */
      saturn.annunc &= ~nibble_masks[addr - 0x10b];
      saturn.annunc |= val << ((addr - 0x10b) * 4);
      display.annunc = saturn.annunc;
      device.ann_touched = 1;
      return;
    case 0x10d:						/* BAUD */
      saturn.baud = val;
      device.baud_touched = 1;
#ifdef DEBUG_SERIAL
      fprintf(stderr, "%.5lx: BAUD write: %x\n", saturn.PC, saturn.baud);
#endif
      return;
    case 0x10e:						/* CARD CONTROL */
      saturn.card_ctrl = val;
      device.card_ctrl_touched = 1;
      return;
    case 0x10f:						/* CARD STATUS */
      saturn.card_status = val;
      device.card_status_touched = 1;
      return;
    case 0x110:						/* IO CONTROL */
      saturn.io_ctrl = val;
      device.ioc_touched = 1;
#ifdef DEBUG_SERIAL
      fprintf(stderr, "%.5lx: IOC write: %x\n", saturn.PC, saturn.io_ctrl);
#endif
      return;
    case 0x111:						/* RCS */
      saturn.rcs = val;
#ifdef DEBUG_SERIAL
      fprintf(stderr, "%.5lx: RCS: %x\n", saturn.PC, saturn.rcs);
#endif
      return;
    case 0x112:						/* TCS */
      saturn.tcs = val;
#ifdef DEBUG_SERIAL
      fprintf(stderr, "%.5lx: TCS: %x\n", saturn.PC, saturn.tcs);
#endif
      return;
    case 0x113:						/* CRER */
      saturn.rcs &= 0x0b;
#ifdef DEBUG_SERIAL
      fprintf(stderr, "%.5lx: CRER, RCS: %x\n", saturn.PC, saturn.rcs);
#endif
      return;
    case 0x114: case 0x115:				/* RBR */
      return;
    case 0x116: case 0x117:				/* TBR */
      saturn.tbr &= ~nibble_masks[addr - 0x116];
      saturn.tbr |= val << ((addr - 0x116) * 4);
      saturn.tcs |= 0x01;
      device.tbr_touched = 1;
      return;
    case 0x118: case 0x119:				/* SERVICE REQ */
      saturn.sreq &= ~nibble_masks[addr - 0x118];
      saturn.sreq |= val << ((addr - 0x118) * 4);
      device.sreq_touched = 1;
#ifdef DEBUG_SERIAL
      fprintf(stderr, "%.5lx: SREQ? write: %x\n", saturn.PC, saturn.sreq);
#endif
      return;
    case 0x11a:						/* IR CONTROL */
      saturn.ir_ctrl = val;
      device.ir_ctrl_touched = 1;
#ifdef DEBUG_IR
      fprintf(stderr, "%.5lx: IRC write: %x\n", saturn.PC, saturn.ir_ctrl);
#endif
      return;
    case 0x11b:						/* BASE NIB OFFSET */
      saturn.base_off = val;
      device.base_off_touched = 1;
      return;
    case 0x11c:						/* LED CONTROL */
      saturn.lcr = val;
      device.lcr_touched = 1;
#ifdef DEBUG_IR
      fprintf(stderr, "%.5lx: LCR write: %x\n", saturn.PC, saturn.lcr);
#endif
      return;
    case 0x11d:						/* LED BUFFER */
      saturn.lbr = val;
      device.lbr_touched = 1;
#ifdef DEBUG_IR
      fprintf(stderr, "%.5lx: LBR write: %x\n", saturn.PC, saturn.lbr);
#endif
      return;
    case 0x11e:						/* SCRATCH PAD */
      saturn.scratch = val;
      device.scratch_touched = 1;
      return;
    case 0x11f:						/* BASENIBBLE */
      saturn.base_nibble = val;
      device.base_nibble_touched = 1;
#ifdef DEBUG_BASE_NIBBLE
      fprintf(stderr, "%.5lx: BASENIB: %x\n", saturn.PC, saturn.base_nibble);
#endif
      return;
    case 0x120: case 0x121: case 0x122: case 0x123:	/* DISP_ADDR */
    case 0x124:
      saturn.disp_addr &= ~nibble_masks[addr - 0x120];
      saturn.disp_addr |= val << ((addr - 0x120) * 4);
      if (display.disp_start != (saturn.disp_addr & 0xffffe)) {
        display.disp_start = saturn.disp_addr & 0xffffe;
        display.disp_end = display.disp_start +
		           (display.nibs_per_line * (display.lines + 1));
        device.display_touched = DISP_INSTR_OFF;
      }
#ifdef DEBUG_DISPLAY
      fprintf(stderr, "%.5lx: DISPLAY: %lx\n", saturn.PC, display.disp_start);
      fprintf(stderr, "%.5lx: DISP END: %lx\n", saturn.PC, display.disp_end);
#endif
      return;
    case 0x125: case 0x126: case 0x127:			/* LINE_OFFSET */
      saturn.line_offset &= ~nibble_masks[addr - 0x125];
      saturn.line_offset |= val << ((addr - 0x125) * 4);
      if (saturn.line_offset != old_line_offset) {
        old_line_offset = saturn.line_offset;
        if (display.offset > 3)
          display.nibs_per_line =
                  (NIBBLES_PER_ROW+saturn.line_offset+2) & 0xfff;
        else
          display.nibs_per_line =
                  (NIBBLES_PER_ROW+saturn.line_offset) & 0xfff;
        display.disp_end = display.disp_start +
		           (display.nibs_per_line * (display.lines + 1));
        device.display_touched = DISP_INSTR_OFF;
      }
#ifdef DEBUG_DISPLAY
      fprintf(stderr, "%.5lx: DISP LINE SIZE: %x\n",
              saturn.PC, display.nibs_per_line);
      fprintf(stderr, "%.5lx: DISP END: %lx\n", saturn.PC, display.disp_end);
#endif
      return;
    case 0x128: case 0x129:				/* LINE_COUNT */
      saturn.line_count &= ~nibble_masks[addr - 0x128];
      saturn.line_count |= val << ((addr - 0x128) * 4);
      line_counter = -1;
      if (display.lines != (saturn.line_count & 0x3f)) {
        display.lines = saturn.line_count & 0x3f;
        disp.lines = 2 * display.lines;
        display.disp_end = display.disp_start +
		           (display.nibs_per_line * (display.lines + 1));
        device.display_touched = DISP_INSTR_OFF;
      }
#ifdef DEBUG_DISPLAY
      fprintf(stderr, "%.5lx: DISP LINES: %x\n", saturn.PC, display.lines);
      fprintf(stderr, "%.5lx: DISP END: %lx\n", saturn.PC, display.disp_end);
#endif
      return;
    case 0x12a: case 0x12b: case 0x12c: case 0x12d:	/* Dont know yet */
      saturn.unknown &= ~nibble_masks[addr - 0x12a];
      saturn.unknown |= val << ((addr - 0x12a) * 4);
#ifdef DEBUG_UNKNOWN 
      fprintf(stderr, "Unknown device @0x%ld: %.4x\n", addr, saturn.unknown);
#endif
      device.unknown_touched = 1;
      return;
    case 0x12e:						/* TIMER 1 CONTROL */
      saturn.t1_ctrl = val;
      device.t1_ctrl_touched = 1;
      return;
    case 0x12f:						/* TIMER 2 CONTROL */
      saturn.t2_ctrl = val;
      device.t2_ctrl_touched = 1;
      return;
    case 0x130: case 0x131: case 0x132: case 0x133:	/* MENU_ADDR */
    case 0x134:
      saturn.menu_addr &= ~nibble_masks[addr - 0x130];
      saturn.menu_addr |= val << ((addr - 0x130) * 4);
      if (display.menu_start != saturn.menu_addr) {
        display.menu_start = saturn.menu_addr;
        display.menu_end = display.menu_start + 0x110;
        device.display_touched = DISP_INSTR_OFF;
      }
      return;
    case 0x135: case 0x136:				/* Dont know yet 2 */
      saturn.unknown2 &= ~nibble_masks[addr - 0x135];
      saturn.unknown2 |= val << ((addr - 0x135) * 4);
#ifdef DEBUG_UNKNOWN 
      fprintf(stderr, "Unknown device @0x%ld: %.2x\n", addr, saturn.unknown2);
#endif
      device.unknown2_touched = 1;
      return;
    case 0x137:						/* TIMER1 */
      saturn.timer1 = val;
      device.t1_touched = 1;
      return;
    case 0x138: case 0x139: case 0x13a: case 0x13b:
    case 0x13c: case 0x13d: case 0x13e: case 0x13f:	/* TIMER2 */
      saturn.timer2 &= ~nibble_masks[addr - 0x138];
      saturn.timer2 |= val << ((addr - 0x138) * 4);
      device.t2_touched = 1;
      return;
    default:
      fprintf(stderr, "%.5lx: UNKNOWN DEVICE WRITE AT 0x%lx !!!\n",
                      saturn.PC, addr);
      return;
  }
}

int
#ifdef __FunctionProto__
read_dev_mem(long addr)
#else
read_dev_mem(addr)
long addr;
#endif
{
  switch (addr) {
    case 0x100:						/* DISPLAY IO */
      return saturn.disp_io & 0x0f;
    case 0x101:						/* CONTRAST CONTROL */
      return saturn.contrast_ctrl & 0x0f;
    case 0x102: case 0x103:				/* DISPLAY TEST */
      return (saturn.disp_test >> ((addr - 0x102) * 4)) & 0x0f;
    case 0x104: case 0x105: case 0x106: case 0x107:	/* CRC */
      return (saturn.crc >> ((addr - 0x104) * 4)) & 0x0f;
    case 0x108:						/* POWER STATUS */
      return saturn.power_status & 0x0f;
    case 0x109:						/* POWER CONTROL */
      return saturn.power_ctrl & 0x0f;
    case 0x10a:						/* MODE */
      return saturn.mode & 0x0f;
    case 0x10b: case 0x10c:				/* ANNUNC */
      return (saturn.annunc >> ((addr - 0x10b) * 4)) & 0x0f;
    case 0x10d:						/* BAUD */
#ifdef DEBUG_SERIAL
      fprintf(stderr, "%.5lx: BAUD read: %x\n", saturn.PC, saturn.baud);
#endif
      return saturn.baud & 0x0f;
    case 0x10e:						/* CARD CONTROL */
      return saturn.card_ctrl & 0x0f;
    case 0x10f:						/* CARD STATUS */
      return saturn.card_status & 0x0f;
    case 0x110:						/* IO CONTROL */
#ifdef DEBUG_SERIAL
      fprintf(stderr, "%.5lx: IOC read: %x\n", saturn.PC, saturn.io_ctrl);
#endif
      return saturn.io_ctrl & 0x0f;
    case 0x111:						/* RCS */
#ifdef DEBUG_SERIAL
      fprintf(stderr, "%.5lx: RCS read: %x\n", saturn.PC, saturn.rcs);
#endif
      return saturn.rcs & 0x0f;
    case 0x112:						/* TCS */
#ifdef DEBUG_SERIAL
      fprintf(stderr, "%.5lx: TCS read: %x\n", saturn.PC, saturn.tcs);
#endif
      return saturn.tcs & 0x0f;
    case 0x113:						/* CRER */
      return 0x00;
    case 0x114: case 0x115:				/* RBR */
      saturn.rcs &= 0x0e;
      device.rbr_touched = 1;
      device_check = 1;
      schedule_event = 0;
      return (saturn.rbr >> ((addr - 0x114) * 4)) & 0x0f;
    case 0x116: case 0x117:				/* TBR */
      return 0x00;
    case 0x118: case 0x119:				/* SERVICE REQ */
#ifdef DEBUG_SERIAL
      fprintf(stderr, "%.5lx: SREQ? read: %x\n", saturn.PC, saturn.sreq);
#endif
      return (saturn.sreq >> ((addr - 0x118) * 4)) & 0x0f;
    case 0x11a:						/* IR CONTROL */
#ifdef DEBUG_IR
      fprintf(stderr, "%.5lx: IRC read: %x\n", saturn.PC, saturn.ir_ctrl);
#endif
      return saturn.ir_ctrl & 0x0f;
    case 0x11b:						/* BASE NIB OFFSET */
      return saturn.base_off & 0x0f;
    case 0x11c:						/* LED CONTROL */
#if 0
#ifdef DEBUG_IR
      fprintf(stderr, "%.5lx: LCR read: %x\n", saturn.PC, saturn.lcr);
#endif
#endif
      return saturn.lcr & 0x0f;
    case 0x11d:						/* LED BUFFER */
#if 0
#ifdef DEBUG_IR
      fprintf(stderr, "%.5lx: LBR read: %x\n", saturn.PC, saturn.lbr);
#endif
#endif
      return saturn.lbr & 0x0f;
    case 0x11e:						/* SCRATCH PAD */
      return saturn.scratch & 0x0f;
    case 0x11f:						/* BASENIBBLE */
      return saturn.base_nibble & 0x0f;
    case 0x120: case 0x121: case 0x122: case 0x123:	/* DISP_ADDR */
    case 0x124:
      return (saturn.disp_addr >> ((addr - 0x120) * 4)) & 0x0f;
    case 0x125: case 0x126: case 0x127:			/* LINE_OFFSET */
      return (saturn.line_offset >> ((addr - 0x125) * 4)) & 0x0f;
    case 0x128:	case 0x129:				/* LINE_COUNT */
      line_counter++;
      if (line_counter > 0x3f)
        line_counter = -1;
      return (((saturn.line_count & 0xc0) | line_counter) >>
             ((addr - 0x128) * 4)) & 0x0f;
    case 0x12a: case 0x12b: case 0x12c: case 0x12d:	/* Dont know yet */
      return (saturn.unknown >> ((addr - 0x12a) * 4)) & 0x0f;
    case 0x12e:						/* TIMER 1 CONTROL */
      return saturn.t1_ctrl & 0x0f;
    case 0x12f:						/* TIMER 2 CONTROL */
      return saturn.t2_ctrl & 0x0f;
    case 0x130: case 0x131: case 0x132: case 0x133:	/* MENU_ADDR */
    case 0x134:
      return (saturn.menu_addr >> ((addr - 0x130) * 4)) & 0x0f;
    case 0x135: case 0x136:				/* Dont know yet 2 */
      return (saturn.unknown2 >> ((addr - 0x135) * 4)) & 0x0f;
    case 0x137:
      return saturn.timer1 & 0xf;
    case 0x138: case 0x139: case 0x13a: case 0x13b:
    case 0x13c: case 0x13d: case 0x13e: case 0x13f:
      return (saturn.timer2 >> ((addr - 0x138) * 4)) & 0xf;
    default:
      fprintf(stderr, "%.5lx: UNKNOWN DEVICE READ AT 0x%lx !!!\n",
                      saturn.PC, addr);
      return 0x00;
  }
}

void
#ifdef __FunctionProto__
write_nibble(long addr, int val)
#else
write_nibble(addr, val)
long addr;
int val;
#endif
{
#if 1
  long orig_addr;
#endif

  addr &= 0xfffff;
#if 1
  orig_addr = addr;
#endif
  val &= 0x0f;
  switch ((addr >> 16) & 0x0f) {
    case 0:
      if (saturn.devices == 0x00100 && addr >= 0x100 && addr < 0x140) {
        write_dev_mem(addr, val);
        return;
      }
#ifdef DEBUG_BAD_MEM
      fprintf(stderr, "%.5lx: BAD WRITE TO ROM AT ADDRESS %.5lx\n",
                      saturn.PC, addr);
#endif
      return;
    case 1: case 2: case 3: case 4: case 5: case 6:
#ifdef DEBUG_BAD_MEM
      fprintf(stderr, "%.5lx: BAD WRITE TO ROM AT ADDRESS %.5lx\n",
                      saturn.PC, addr);
#endif
      return;
    case 7:
      switch (saturn.ram32k) {
        case 0x70000:
          break;
        case 0xfc000:
          if (addr < 0x74000)
            break;
          goto DEFAULT;
        case 0xfe000:
          if (addr < 0x72000)
            break;
        default:
        DEFAULT:
#ifdef DEBUG_BAD_MEM
          fprintf(stderr, "%.5lx: BAD WRITE TO ROM AT ADDRESS %.5lx\n",
                          saturn.PC, addr);
#endif
          return;
      }
      saturn.ram[addr - 0x70000] = val;
      break;
    case 8: case 9: case 0xa: case 0xb:
      saturn.port1[addr - 0x80000] = val;
      return;
    case 0xc: case 0xd: case 0xe:
      saturn.port2[addr - 0xc0000] = val;
      return;
    case 0xf:
      switch (saturn.ram32k) {
        case 0x70000:
          saturn.port2[addr - 0xc0000] = val;
          return;
        case 0xfc000:
          if (addr >= 0xfc000) {
            saturn.ram[addr - 0xfc000] = val;
            break;
          }
          saturn.port2[addr - 0xc0000] = val;
          return;
        case 0xfe000:
          if (addr >= 0xfe000) {
            saturn.ram[addr - 0xfe000] = val;
            break;
          }
          saturn.port2[addr - 0xc0000] = val;
          return;
        case 0xf0000:
          saturn.ram[addr - 0xf0000] = val;
          break;
        default:
#ifdef DEBUG_BAD_MEM
          fprintf(stderr, "%.5lx: BAD WRITE TO RAM AT ADDRESS %.5lx\n",
                           saturn.PC, addr);
#endif
          return;
      }
      break;
  }
#if 1
  if (device.display_touched || !disp.mapped) return;
  if (orig_addr >= display.disp_start && orig_addr < display.disp_end) {
    disp_draw_nibble(orig_addr, val);
  }
  if (display.lines == 63) return;
  if (orig_addr >= display.menu_start && orig_addr < display.menu_end) {
    menu_draw_nibble(orig_addr, val);
  }
#endif
}

int
#ifdef __FunctionProto__
read_nibble(long addr)
#else
read_nibble(addr)
long addr;
#endif
{
  addr &= 0xfffff;
  switch ((addr >> 16) & 0x0f) {
    case 0:
      if (addr >= 0x100 && addr < 0x140) {
        if (saturn.devices == 0x00100) {
          return read_dev_mem(addr);
        } else {
          return 0x00;
        }
      }
      return saturn.rom[addr] & 0x0f;
    case 1: case 2: case 3: case 4: case 5: case 6:
      return saturn.rom[addr] & 0x0f;
    case 7:
      switch (saturn.ram32k) {
        case 0x70000:
          return saturn.ram[addr - 0x70000];
        case 0xf0000:
          return saturn.rom[addr] & 0x0f;
        case 0xfc000:
          if (addr < 0x74000) {
            return saturn.ram[addr - 0x70000];
          }
          return saturn.rom[addr] & 0x0f;
        case 0xfe000:
          if (addr < 0x72000) {
            return saturn.ram[addr - 0x70000];
          }
          return saturn.rom[addr] & 0x0f;
        default:
#ifdef DEBUG_BAD_MEM
          fprintf(stderr, "%.5lx: ATTEMPT TO READ FROM %.5lx\n",
                  saturn.PC, addr);
#endif
          return 0x00;
      }
    case 8: case 9: case 0xa: case 0xb:
      return saturn.port1[addr - 0x80000] & 0x0f;
    case 0xc: case 0xd: case 0xe:
      return saturn.port2[addr - 0xc0000] & 0x0f;
    case 0xf:
      switch (saturn.ram32k) {
        case 0x70000:
          return saturn.port2[addr - 0xc0000] & 0x0f;
        case 0xfc000:
          if (addr >= 0xfc000) {
            return saturn.ram[addr - 0xfc000];
          }
          return saturn.port2[addr - 0xc0000] & 0x0f;
        case 0xfe000:
          if (addr >= 0xfe000) {
            return saturn.ram[addr - 0xfe000];
          }
          return saturn.port2[addr - 0xc0000] & 0x0f;
        case 0xf0000:
          return saturn.ram[addr - 0xf0000];
        default:
#ifdef DEBUG_BAD_MEM
          fprintf(stderr, "%.5lx: ATTEMPT TO READ FROM %.5lx\n",
                  saturn.PC, addr);
#endif
          return 0x00;
      }
    default:
#ifdef DEBUG_BAD_MEM
      fprintf(stderr, "%.5lx: ATTEMPT TO READ FROM %.5lx\n",
                      saturn.PC, addr);
#endif
      return 0x00;
  }
}

int
#ifdef __FunctionProto__
read_nibble_crc(long addr)
#else
read_nibble_crc(addr)
long addr;
#endif
{
  addr &= 0xfffff;
  switch ((addr >> 16) & 0x0f) {
    case 0:
      if (addr >= 0x100 && addr < 0x140) {
        if (saturn.devices == 0x00100) {
          return read_dev_mem(addr);
        } else {
          return calc_crc(0x00);
        }
      }
      return calc_crc(saturn.rom[addr] & 0x0f);
    case 1: case 2: case 3: case 4: case 5: case 6:
      return calc_crc(saturn.rom[addr] & 0x0f);
    case 7:
      switch (saturn.ram32k) {
        case 0x70000:
          return calc_crc(saturn.ram[addr - 0x70000]);
        case 0xf0000:
          return calc_crc(saturn.rom[addr] & 0x0f);
        case 0xfc000:
          if (addr < 0x74000) {
            return calc_crc(saturn.ram[addr - 0x70000]);
          }
          return calc_crc(saturn.rom[addr] & 0x0f);
        case 0xfe000:
          if (addr < 0x72000) {
            return calc_crc(saturn.ram[addr - 0x70000]);
          }
          return calc_crc(saturn.rom[addr] & 0x0f);
        default:
#ifdef DEBUG_BAD_MEM
          fprintf(stderr, "%.5lx: ATTEMPT TO READ FROM %.5lx\n",
                  saturn.PC, addr);
#endif
          return 0x00;
      }
    case 8: case 9: case 0xa: case 0xb:
      return calc_crc(saturn.port1[addr - 0x80000] & 0x0f);
    case 0xc: case 0xd: case 0xe:
      return calc_crc(saturn.port2[addr - 0xc0000] & 0x0f);
    case 0xf:
      switch (saturn.ram32k) {
        case 0x70000:
          return calc_crc(saturn.port2[addr - 0xc0000] & 0x0f);
        case 0xfc000:
          if (addr >= 0xfc000) {
            return calc_crc(saturn.ram[addr - 0xfc000]);
          }
          return calc_crc(saturn.port2[addr - 0xc0000] & 0x0f);
        case 0xfe000:
          if (addr >= 0xfe000) {
            return calc_crc(saturn.ram[addr - 0xfe000]);
          }
          return calc_crc(saturn.port2[addr - 0xc0000] & 0x0f);
        case 0xf0000:
          return calc_crc(saturn.ram[addr - 0xf0000]);
        default:
#ifdef DEBUG_BAD_MEM
          fprintf(stderr, "%.5lx: ATTEMPT TO READ FROM %.5lx\n",
                  saturn.PC, addr);
#endif
          return 0x00;
      }
    default:
#ifdef DEBUG_BAD_MEM
      fprintf(stderr, "%.5lx: ATTEMPT TO READ FROM %.5lx\n",
                      saturn.PC, addr);
#endif
      return 0x00;
  }
}

long
#ifdef __FunctionProto__
read_nibbles(long addr, int len)
#else
read_nibbles(addr, len)
long addr;
int len;
#endif
{
  long val = 0;

  addr += len;
  while (len-- > 0) {
    val = (val << 4) | read_nibble(--addr);
  }
  return val;
}

void
#ifdef __FunctionProto__
write_nibbles(long addr, long val, int len)
#else
write_nibbles(addr, val, len)
long addr;
long val;
int len;
#endif
{
  while (len-- > 0) {
    write_nibble(addr++, val);
    val >>= 4;
  }
}

void
#ifdef __FunctionProto__
dev_memory_init(void)
#else
dev_memory_init()
#endif
{
  write_nibble(0x10f, 0xf);	/* CARDSTAT */
  memset(&device, 0, sizeof(device));
  init_display();
}
